from wp.command import WPCommand


# Flushes the object cache.
# For WordPress multisite instances using a persistent object cache,
# flushing the object cache will typically flush the cache for all sites.
# Beware of the performance impact when flushing the object cache in production.
# Errors if the object cache can’t be flushed.
# wp cache flush
# Success: The cache was flushed.
class CacheFlush(WPCommand):
    command = ['cache', 'flush']

    def __init__(self, **args):
        super().__init__(**args)

    def params(self):
        return []



from wp.cache_set import CacheSet


# TODO Test.
# Adds a value to the object cache.
# Errors if a value already exists for the key, which means the value can’t be added.
# wp cache add my_key my_group my_value 300
# Success: Added object 'my_key' in group 'my_value'.
class CacheAdd(CacheSet):
    command = ['cache', 'add']



from wp.cache_get import CacheGet


# TODO Test.
# Removes a value from the object cache.
# Errors if the value can’t be deleted.
# wp cache delete my_key my_group
# Success: Object deleted.
class CacheDelete(CacheGet):
    command = ['cache', 'delete']




from wp.command import WPCommand


# Attempts to determine which object cache is being used.
# Note that the guesses made by this function are based on the WP_Object_Cache classes
# that define the 3rd party object cache extension. Changes to those classes could render
# problems with this function’s ability to determine which object cache is being used.
# wp cache type
# Default
class CacheType(WPCommand):
    command = ['cache', 'type']

    def __init__(self, **args):
        super().__init__(**args)

    def params(self):
        return []



from wp.cache_set import CacheSet


# TODO Test.
# Replaces a value in the object cache, if the value already exists.
# Errors if the value can’t be replaced.
# wp cache replace my_key new_value my_group
# Success: Replaced object 'my_key' in group 'my_group'.
class CacheReplace(CacheSet):
    command = ['cache', 'replace']



from wp import DBTables


class DBItem(object):
    # From Search
    table = ''
    column = ''
    id = int
    value = ''

    # Column of the ID.
    id_column = ''

    @staticmethod
    def new_from_db_table_entry(table, entry):
        item = DBItem()
        item.entry = entry
        item.table = table

        return item

    @staticmethod
    def new_from_db_search_line(line):
        return DBItem.new_from_db_search_item(line.split(":", 3))

    @staticmethod
    def new_from_db_search_item(parts):
        item = DBItem()
        if type(parts) == list and len(parts) == 4:
            item.table = parts[0]
            item.column = parts[1]
            item.id = int(parts[2])
            item.value = parts[3]

        return item


class DBItems(list):
    __all_tables = []
    lines = []
    items = []
    tables = []

    def __init__(self):
        super().__init__()
        self.set_all_tables()

    @staticmethod
    def new_from_search_output(search_output_str):
        items = DBItems()

        items.lines = search_output_str.splitlines()
        for index, line in enumerate(items.lines):
            if line.startswith(tuple(items.__all_tables)):
                item = DBItem.new_from_db_search_line(line)
                items.items.append(item)
            else:
                items.items[-1].value = "\n".join([items.items[-1].value, line])

        return items

    def get_all_items_from_table(self, table):
        return [a for a in self.items if table == a.table]

    def set_all_tables(self):
        if 0 != len(self.__all_tables):
            return self.__all_tables

        tables = DBTables(all_tables=True)
        out, err = tables.run()
        self.__all_tables = tables.tables.get_tables()



from wp.command import WPCommand


class Tables(object):
    tables = []

    def __init__(self, cmd_output, format="list"):
        if "csv" == format:
            self.tables = cmd_output.split(",")
        else:
            self.tables = cmd_output.split("\n")

    def has_table(self, table):
        return table in self.tables

    def get_tables(self):
        return self.tables


class DBTables(WPCommand):
    command = ['db', 'tables']

    # List tables based on wildcard search, e.g. ‘wp_*_options’ or ‘wp_post?’.
    table = ''

    # Can be all, global, ms_global, blog, or old tables. Defaults to all.
    scope = ''
    __scope_enum = [
        'all',
        'global',
        'ms_global',
        'blog',
        'old',
    ]

    # List all the tables in a multisite install.
    network = bool

    # List all tables that match the table prefix even if not registered on $wpdb. Overrides –network.
    all_tables_with_prefix = bool

    # List all tables in the database, regardless of the prefix, and even if not registered on $wpdb.
    # Overrides –all-tables-with-prefix.
    all_tables = bool

    # Render output in a particular format.
    # default: list
    # options: list, csv
    format = ''

    # Tables from the command.
    tables = Tables

    def __init__(self, **args):
        super().__init__(**args)

        self.table = self.get_arg_value(key="table", default_value=self.table)
        self.scope = self.get_arg_value(key="scope", default_value=self.scope)
        self.network = self.get_arg_value(key="network", default_value=self.network)
        self.all_tables_with_prefix = self.get_arg_value(key="all_tables_with_prefix",
                                                         default_value=self.all_tables_with_prefix)
        self.all_tables = self.get_arg_value(key="all_tables", default_value=self.all_tables)
        self.format = self.get_arg_value(key="format", default_value=self.format)

    def params(self):
        return []

    def run(self):
        out, err = super().run()
        self.tables = Tables(out)
        return out, err



from wp.command import WPCommand


class DBCheck(WPCommand):
    command = ['db', 'check']

    # Extra arguments to pass to mysqldump. Refer to mysqldump docs.
    fields = []

    # Loads the environment’s MySQL option files.
    # Default behavior is to skip loading them to avoid failures due to misconfiguration.
    defaults = []

    def __init__(self, **args):
        super().__init__(**args)

        self.fields = self.get_arg_value(key="fields", default_value=self.fields)
        self.defaults = self.get_arg_value(key="defaults", default_value=self.defaults)

    def params(self):
        return []



import os

from loggr import logger
import subprocess


class WPCommand(object):
    # Holder for the args.
    _args = {}

    # User running command.
    cmd_user = "www-data"

    # Whether to run as sudo.
    as_sudo = False
    __as_sudo = ["sudo", "-u", cmd_user]
    __as_sudo_param = ["--allow-root"]

    # What should come before the WP Command.
    cmd_prefix = []

    # WordPress command binary
    cmd_base = 'wp'

    # What should follow the WP Command.
    cmd_suffix = []

    # The WP Command.
    command = []

    # GLOBAL PARAMETERS.

    # Path to the WordPress files.
    path = ''

    # Perform operation against a remote server over SSH.
    ssh = ''

    # Perform operation against a remote WordPress install over HTTP.
    http = ''

    # Pretend request came from given URL. In multisite, this argument is how the target site is specified.
    url = ''

    # Set the WordPress user. Options: id,email,login
    user = ''

    # Skip loading all or some plugins. Note: mu-plugins are still loaded.
    skip_plugins = ''

    # Skip loading all or some themes.
    skip_themes = ''

    # Skip loading all installed packages.
    skip_packages = []

    # Whether to colorize the output.
    color = ''

    # Show all PHP errors; add verbosity to WP-CLI bootstrap.
    # Not implemented
    debug = bool

    # Suppress informational messages.
    # Not implemented.
    quiet = False

    # List of Apache Modules that are to be reported as loaded.
    apache_modules = []

    # List of PHP files to require before running.
    require = []

    # PHP to execute
    exec = ''

    # Whether to prompt
    prompt = False

    # Environment Variables.
    #  TODO Make this do something.
    envs = {
        # WP_CLI_CACHE_DIR – Directory to store the WP-CLI file cache.
        # Default is ~/.wp-cli/cache/.
        "WP_CLI_CACHE_DIR": '~/.wp-cli/cache/',

        # WP_CLI_CONFIG_PATH – Path to the global config.yml file.
        # Default is ~/.wp-cli/config.yml.
        "WP_CLI_CONFIG_PATH": '~/.wp-cli/config.yml',

        # WP_CLI_CUSTOM_SHELL – Allows the user to override the default /bin/bash shell used.
        "WP_CLI_CUSTOM_SHELL": '',

        # WP_CLI_DISABLE_AUTO_CHECK_UPDATE – Disable WP-CLI automatic checks for updates.
        "WP_CLI_DISABLE_AUTO_CHECK_UPDATE": '',

        # WP_CLI_PACKAGES_DIR – Directory to store packages installed through WP-CLI’s package management.
        # Default is ~/.wp-cli/packages/.
        "WP_CLI_PACKAGES_DIR": '~/.wp-cli/packages/',

        # WP_CLI_PHP – PHP binary path to use when overriding the system default (only works for non-Phar installation).
        "WP_CLI_PHP": '',

        # WP_CLI_PHP_ARGS – Arguments to pass to the PHP binary when invoking WP-CLI (only works for non-Phar installation).
        "WP_CLI_PHP_ARGS": '',

        # WP_CLI_SSH_PRE_CMD – When using --ssh:<ssh>, perform a command before WP-CLI calls WP-CLI on the remote server.
        "WP_CLI_SSH_PRE_CMD": '',

        # WP_CLI_STRICT_ARGS_MODE – Avoid ambiguity by telling WP-CLI to treat any arguments before the command as global,
        # and after the command as local.
        "WP_CLI_STRICT_ARGS_MODE": '',

        # WP_CLI_SUPPRESS_GLOBAL_PARAMS – Set to true to skip showing the global parameters at the end of the help screen.
        # This saves screen estate for advanced users.
        "WP_CLI_SUPPRESS_GLOBAL_PARAMS": '',
    }

    # STDOUT result of run().
    output = ''

    # STDERR result of run().
    error = ''

    # def __setattr__(key, value):
    #     super().__setattr__(key, value)

    def __init__(self, **args):
        super().__init__()
        if self._args is not None:
            self._args = args

        self.cmd_prefix = self.get_arg_value(key="cmd_prefix", default_value=self.cmd_prefix)
        self.as_sudo = self.get_arg_value(key="as_sudo", default_value=self.as_sudo)
        self.cmd_user = self.get_arg_value(key="cmd_user", default_value=self.cmd_user)

        # Global parameters.
        self.path = self.get_arg_value(key="path", default_value=self.path)
        self.ssh = self.get_arg_value(key="ssh", default_value=self.ssh)
        self.http = self.get_arg_value(key="http", default_value=self.http)
        self.url = self.get_arg_value(key="url", default_value=self.url)
        self.user = self.get_arg_value(key="user", default_value=self.user)
        self.skip_plugins = self.get_arg_value(key="skip_plugins", default_value=self.skip_plugins)
        self.skip_themes = self.get_arg_value(key="skip_themes", default_value=self.skip_themes)
        self.skip_packages = self.get_arg_value(key="skip_packages", default_value=self.skip_packages)
        self.color = self.get_arg_value(key="color", default_value=self.color)
        self.debug = self.get_arg_value(key="debug", default_value=self.debug)
        # self.quiet = self.get_arg_value(key="quiet", default_value=self.quiet)
        self.apache_modules = self.get_arg_value(key="apache_modules", default_value=self.apache_modules)

    def __getitem__(self, item):
        return super().__getattribute__(item)

    def __str__(self):
        return " ".join([str(elem) for elem in self.cmd()])

    # PUBLIC API.
    def get_arg_value(self, key, default_value):
        value = self._args.get(key)
        if value is None:
            return default_value

        return value

    # Gets everything for the command.
    def cmd(self):
        return self.__cmd()

    def __cmd(self):
        return self.__prefix() + \
               self.__wp_cmd() + \
               self.command + \
               self.__params() + \
               self.__suffix()

    # Runs the command as sudo or as root.
    def run_as_sudo(self):
        self.as_sudo = True
        return self.run()

    # Runs the command.
    def run(self):
        c = self.cmd()
        logger.info(f"running " + " ".join([str(elem) for elem in c]))
        # print(f"running: " + " ".join([str(elem) for elem in c]))

        output, error = subprocess.Popen(
            c, universal_newlines=True,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()

        self.output = output.strip("\n")
        self.error = error.strip("\n")

        # print(output)
        # print(error)
        logger.info("Output: {0}".format(self.output))
        logger.info("Errors: {0}".format(self.error))
        return self.output, self.error

        # output = subprocess.check_output(c)
        # output = subprocess.check_output(c, stderr=subprocess.STDOUT)
        # print(output.decode('utf-8'))

        # try:
        #     output = check_output(c, stderr=STDOUT)
        #     # if "This does not seem to be a WordPress installation"
        #     logger.info(output)
        # except CalledProcessError as e:
        #     o = e.output.decode("utf-8")
        #     print(o)
        #     return o

    # Abstract Method. Returns the parameters.
    def params(self):
        raise NotImplementedError("Please implement params")

    # Abstract Method. Returns the parsed output or by default, just the output.
    def get_result(self):
        return self.output

    # Params that should not have underscores (_) replaced with dashes (-).
    # Should be overridden.
    def get_raw_params(self):
        return []

    # Gets excluded attributes. These class/instance attributes will never be outputted.
    # Should be overridden.
    def get_excluded_attrs(self):
        return []

    # Gets the bool attributes. Key should be the raw attribute.
    def get_custom_bool_attrs(self):
        return {}

    # Returns a custom string output for specific attribute.
    def get_attr_custom_param_output(self, attr):
        return ''

    # PRIVATE API.

    # Params that should not have underscores (_) replaced with dashes (-).
    # Should be overridden.
    def __get_raw_params(self):
        return self.get_raw_params() + [
            "apache_modules",
        ]

    # Gets the custom bool attrs. Key should be the raw attribute.
    def __get_custom_bool_attrs(self):
        global_attrs = {
            "color-false": "--no-color",
            "debug-false": "--debug=false",
        }

        attrs = self.get_custom_bool_attrs()
        attrs.update(global_attrs)
        return attrs

    # Gets the custom string attrs.
    def __get_attr_custom_param_output(self, attr):
        if 'format' == attr and hasattr(self, 'format') and '' != self.format:
            return '--format=' + self.format

        return self.get_attr_custom_param_output(attr)

    # Gets excluded attributes.
    def __get_excluded_attrs(self):
        return self.get_excluded_attrs() + [
            "_args",
            "as_sudo",
            "cmd_user",
            "cmd_prefix",
            "cmd_base",
            "cmd_suffix",
            "command",
            "envs",
            "error",
            "output",
            "prompt",
            "quiet",
        ]

    def __filter_atts(self, attr):
        # res = (
        #         not attr.startswith('__')
        #         and not attr.startswith('_' + self.__class__.__name__ + '__')
        #         and not attr.startswith('_WPCommand__')
        #         and not callable(getattr(self, attr))
        #         and attr not in self.__get_excluded_attrs()
        # )
        return (
                not attr.startswith('__')
                and not attr.startswith('_' + self.__class__.__name__ + '__')
                and not attr.startswith('_WPCommand__')
                and not callable(getattr(self, attr))
                and attr not in self.__get_excluded_attrs()
        )

    # Gets all attributes from class that are command parameters
    def __get_attrs(self):
        # print(dir(self))
        orig_attrs = dir(self)
        # attrs = list(filter(self.__filter_atts, dir(self)))
        attrs = [a for a in dir(self) if self.__filter_atts(a)]
        return attrs

    # Returns the prefix.
    def __prefix(self):
        return self.cmd_prefix

    # Returns the user running the command and the command base (`wp`).
    def __wp_cmd(self):
        if self.as_sudo:
            return self.__as_sudo + [
                self.cmd_base
            ]
        return [
            self.cmd_base
        ]

    # Returns the parameters.
    def __params(self):
        # return self._get_params_by_attrs(self._get_attrs(), self.params())
        p = self.params()
        attrs = self.__get_attrs()
        for attr in attrs:
            key = attr
            val = self[attr]
            if attr not in self.__get_raw_params():
                key = key.replace('_', '-')

            custom_out_attr = self.__get_attr_custom_param_output(attr)
            if '' != custom_out_attr:
                p.append(custom_out_attr)
                continue

            t = type(self[attr])
            if bool == t:
                custom_bool_attrs = self.__get_custom_bool_attrs()
                if f"{attr}-{str(val).lower()}" in custom_bool_attrs.keys():
                    p.append(custom_bool_attrs[f"{attr}-{str(val).lower()}"])
                elif self[attr]:
                    p.append(f"--{key}")
            elif list == t:
                if 0 != len(val):
                    val_str = ",".join(val)
                    p.append(f"--{key}=\"{val_str}\"")
            elif str == t:
                if '' != val:
                    # p.append(f"--{key}={val}")
                    p.append(f"--{key}=\"{val}\"")
            else:
                p.append(f"--{key}=\"{val}\"")

        if self.as_sudo and ('root' == self.cmd_user or os.geteuid() == 0):
            p += self.__as_sudo_param

        return p

    def __suffix(self):
        return self.cmd_suffix



from datetime import datetime

from wp.command import WPCommand


class PostUpdate(WPCommand):
    command = ['post', 'update']

    # Post ID.
    id = int

    # [--post_author=<post_author>]
    # The ID of the user who added the post. Default is the current user ID.
    post_author = int

    # [--post_date=<post_date>]
    # The date of the post. Default is the current time.
    post_date = datetime

    # [--post_date_gmt=<post_date_gmt>]
    # The date of the post in the GMT timezone. Default is the value of $post_date.
    post_date_gmt = datetime

    # [--post_content=<post_content>]
    # The post content. Default empty.
    post_content = ''

    # [--post_content_filtered=<post_content_filtered>]
    # The filtered post content. Default empty.
    post_content_filtered = ''

    # [--post_title=<post_title>]
    # The post title. Default empty.
    post_title = ''

    # [--post_excerpt=<post_excerpt>]
    # The post excerpt. Default empty.
    post_excerpt = ''

    # [--post_status=<post_status>]
    # The post status. Default ‘draft’.
    post_status = ''

    # [--post_type=<post_type>]
    # The post type. Default ‘post’.
    post_type = ''

    # [--comment_status=<comment_status>]
    # Whether the post can accept comments. Accepts ‘open’ or ‘closed’.
    # Default is the value of ‘default_comment_status’ option.
    comment_status = ''

    # [--ping_status=<ping_status>]
    # Whether the post can accept pings. Accepts ‘open’ or ‘closed’.
    # Default is the value of ‘default_ping_status’ option.
    ping_status = ''

    # [--post_password=<post_password>]
    # The password to access the post. Default empty.
    post_password = ''

    # [--post_name=<post_name>]
    # The post name. Default is the sanitized post title when creating a new post.
    post_name = ''

    # [--to_ping=<to_ping>]
    # Space or carriage return-separated list of URLs to ping. Default empty.
    to_ping = ''

    # [--pinged=<pinged>]
    # Space or carriage return-separated list of URLs that have been pinged. Default empty.
    pinged = ''

    # [--post_modified=<post_modified>]
    # The date when the post was last modified. Default is the current time.
    post_modified = datetime

    # [--post_modified_gmt=<post_modified_gmt>]
    # The date when the post was last modified in the GMT timezone. Default is the current time.
    post_modified_gmt = datetime

    # [--post_parent=<post_parent>]
    # Set this for the post it belongs to, if any. Default 0.
    post_parent = int

    # [--menu_order=<menu_order>]
    # The order the post should be displayed in. Default 0.
    menu_order = int

    # [--post_mime_type=<post_mime_type>]
    # The mime type of the post. Default empty.
    post_mime_type = ''

    # [--guid=<guid>]
    # Global Unique ID for referencing the post. Default empty.
    # GUID cannot be updated, so why even support it?
    # guid = ''

    # [--post_category=<post_category>]
    # Array of category names, slugs, or IDs. Defaults to value of the ‘default_category’ option.
    post_category = ''

    # [--tags_input=<tags_input>]
    # Array of tag names, slugs, or IDs. Default empty.
    tags_input = ''

    # [--tax_input=<tax_input>]
    # Array of taxonomy terms keyed by their taxonomy name. Default empty.
    tax_input = ''

    # [--meta_input=<meta_input>]
    # Array in JSON format of post meta values keyed by their post meta key. Default empty.
    meta_input = ''

    # [<file>]
    # Read post content from <file>. If this value is present, the
    # --post_content argument will be ignored.
    # Passing - as the filename will cause post content to
    # be read from STDIN.
    file = ''

    # --<field>=<value>
    # One or more fields to update. See wp_insert_post().
    field = ''

    # [--defer-term-counting]
    # Recalculate term count in batch, for a performance boost.
    # Instead of returning the whole post, returns the value of a single field.
    defer_term_counting = ''

    def __init__(self, id, **args):
        super().__init__(**args)
        self.id = id

        self.post_author = self.get_arg_value(key="post_author", default_value=self.post_author)
        self.post_date = self.get_arg_value(key="post_date", default_value=self.post_date)
        self.post_date_gmt = self.get_arg_value(key="post_date_gmt", default_value=self.post_date_gmt)
        self.post_content = self.get_arg_value(key="post_content", default_value=self.post_content)
        self.post_content_filtered = self.get_arg_value(key="post_content_filtered", default_value=self.post_content_filtered)
        self.post_title = self.get_arg_value(key="post_title", default_value=self.post_title)
        self.post_excerpt = self.get_arg_value(key="post_excerpt", default_value=self.post_excerpt)
        self.post_status = self.get_arg_value(key="post_status", default_value=self.post_status)
        self.post_type = self.get_arg_value(key="post_type", default_value=self.post_type)
        self.comment_status = self.get_arg_value(key="comment_status", default_value=self.comment_status)
        self.ping_status = self.get_arg_value(key="ping_status", default_value=self.ping_status)
        self.post_password = self.get_arg_value(key="post_password", default_value=self.post_password)
        self.post_name = self.get_arg_value(key="post_name", default_value=self.post_name)
        self.to_ping = self.get_arg_value(key="to_ping", default_value=self.to_ping)
        self.post_modified = self.get_arg_value(key="post_modified", default_value=self.post_modified)
        self.post_modified_gmt = self.get_arg_value(key="post_modified_gmt", default_value=self.post_modified_gmt)
        self.post_parent = self.get_arg_value(key="post_parent", default_value=self.post_parent)
        self.menu_order = self.get_arg_value(key="menu_order", default_value=self.menu_order)
        self.post_mime_type = self.get_arg_value(key="post_mime_type", default_value=self.post_mime_type)
        # self.guid = self.get_arg_value(key="guid", default_value=self.guid)
        self.post_category = self.get_arg_value(key="post_category", default_value=self.post_category)
        self.tags_input = self.get_arg_value(key="tags_input", default_value=self.tags_input)
        self.tax_input = self.get_arg_value(key="tax_input", default_value=self.tax_input)
        self.meta_input = self.get_arg_value(key="meta_input", default_value=self.meta_input)
        self.field = self.get_arg_value(key="field", default_value=self.field)
        self.defer_term_counting = self.get_arg_value(key="defer_term_counting", default_value=self.defer_term_counting)

    def params(self):
        return [
            str(self.id)
        ]

    def get_excluded_attrs(self):
        return [
            "id"
        ]

    def get_raw_params(self):
        return [
            'post_author',
            'post_date',
            'post_date_gmt',
            'post_content',
            'post_content_filtered',
            'post_title',
            'post_excerpt',
            'post_status',
            'post_type',
            'comment_status',
            'ping_status',
            'post_password',
            'post_name',
            'to_ping',
            'post_modified',
            'post_modified_gmt',
            'post_parent',
            'menu_order',
            'post_mime_type',
            'post_category',
            'tags_input',
            'tax_input',
            'meta_input',
        ]



from wp.command import WPCommand


class DBSearch(WPCommand):
    command = ['db', 'search']

    # SQL.
    search = ''  # Required

    # One or more tables to search through for the string.
    tables = ''

    # Search through all the tables registered to $wpdb in a multisite install.
    network = bool

    # Enable replacement on any tables that match the table prefix even if not registered on $wpdb.
    all_tables_with_prefix = bool

    # Enable replacement on ALL tables in the database, regardless of the prefix, and even if not registered on $wpdb.
    # Overrides –network and –all-tables-with-prefix.
    all_tables = bool

    # For logging, number of characters to display before the old match and the new replacement.
    # Default 40. Ignored if not logging.
    before_context = int

    # For logging, number of characters to display after the old match and the new replacement.
    # Default 40. Ignored if not logging.
    after_context = int

    # Runs the search as a regular expression (without delimiters). The search becomes case-sensitive
    # (i.e. no PCRE flags are added). Delimiters must be escaped if they occur in the expression.
    # Because the search is run on individual columns, you can use the ^ and $ tokens
    # to mark the start and end of a match, respectively.
    regex = bool

    # Pass PCRE modifiers to the regex search (e.g. ‘i’ for case-insensitivity).
    regex_flags = ''

    # The delimiter to use for the regex. It must be escaped if it appears in the search string.
    # The default value is the result of chr(1).
    regex_delimiter = ''

    # Output the ‘table:column’ line once before all matching row lines in the table column
    # rather than before each matching row.
    table_column_once = bool

    # Place the ‘table:column’ output on the same line as the row id and match (‘table:column:id:match’).
    # Overrides –table_column_once.
    one_line = bool

    # Only output the string matches (including context). No ‘table:column’s or row ids are outputted.
    matches_only = bool

    # Output stats on the number of matches found, time taken, tables/columns/rows searched, tables skipped.
    stats = bool

    # Percent color code to use for the ‘table:column’ output. For a list of available percent color codes, see below.
    # Default ‘%G’ (bright green).
    table_column_color = ''

    # Percent color code to use for the row id output. For a list of available percent color codes, see below.
    # Default ‘%Y’ (bright yellow).
    id_color = ''

    # Percent color code to use for the match (unless both before and after context are 0, when no color code is used).
    # For a list of available percent color codes, see below. Default ‘%3%k’ (black on a mustard background).
    match_color = ''

    def __init__(self, search, **args):
        super().__init__(**args)
        self.search = search

        self.all_tables_with_prefix = self.get_arg_value(key="all_tables_with_prefix",
                                                         default_value=self.all_tables_with_prefix)
        self.all_tables = self.get_arg_value(key="all_tables", default_value=self.all_tables)
        self.before_context = self.get_arg_value(key="before_context", default_value=self.before_context)
        self.after_context = self.get_arg_value(key="after_context", default_value=self.after_context)
        self.regex = self.get_arg_value(key="regex", default_value=self.regex)
        self.regex_flags = self.get_arg_value(key="regex_flags", default_value=self.regex_flags)
        self.regex_delimiter = self.get_arg_value(key="regex_delimiter", default_value=self.regex_delimiter)
        self.table_column_once = self.get_arg_value(key="table_column_once", default_value=self.table_column_once)
        self.one_line = self.get_arg_value(key="one_line", default_value=self.one_line)
        self.matches_only = self.get_arg_value(key="matches_only", default_value=self.matches_only)
        self.stats = self.get_arg_value(key="stats", default_value=self.stats)
        self.table_column_color = self.get_arg_value(key="table_column_color", default_value=self.table_column_color)
        self.id_color = self.get_arg_value(key="id_color", default_value=self.id_color)
        self.match_color = self.get_arg_value(key="match_color", default_value=self.match_color)

    def params(self):
        return [
            self.search
        ]

    def get_raw_params(self):
        return [
            "before_context",
            "after_context",
            "table_column_once",
            "one_line",
            "matches_only",
            "table_column_color",
            "id_color",
            "match_color",
        ]

    def get_excluded_attrs(self):
        return [
            "search"
        ]



from wp.cap import Cap


# TODO Test.
# Adds capabilities to a given role.
# wp cap add author spectate
# Success: Added 1 capability to 'author' role.
class CapAdd(Cap):
    command = ['cap', "add"]

    # <role>
    # Key for the role.
    role = ""

    # <cap>…
    # One or more capabilities to add.
    cap = ""

    # [--grant]
    # Adds the capability as an explicit boolean value, instead of implicitly defaulting to true.
    # ---
    # default: true
    # options:
    # – true
    # – false
    # ---

    def __init__(self, key, value, **args):
        super().__init__(key, **args)

        self.value = value
        self.expiration = self.get_arg_value(key="expiration", default_value=self.expiration)

    def params(self):
        return [
            self.key,
            self.value,
            self.group,
            self.expiration,
        ]

    def get_excluded_attrs(self):
        return super().get_excluded_attrs() + [
            "value",
            "expiration",
        ]



import json
from wp.command import WPCommand


class Column(object):
    field = ''
    type = ''
    null = bool
    key = ''
    default = ''
    extra = ''

    @staticmethod
    def new_from_dict(d):
        if type(d) != dict:
            raise TypeError("column needs to be a dictionary")

        column = Column()
        column.field = d['Field']
        column.type = d['Type']
        column.null = bool(d['Null'].lower())
        column.key = d['Key']
        column.default = d['Default']
        column.extra = d['Extra']

        return column

    @staticmethod
    def new_from_list(l):
        if type(l) != list and 6 > len(l):
            raise TypeError("column needs to be a list of 6 elements")

        column = Column()
        column.field = l[0]
        column.type = l[1]
        column.null = bool(l[2].lower())
        column.key = l[3]
        column.default = l[4]
        column.extra = l[5]

        return column


class Columns(object):
    columns = {}

    def __init__(self, cmd_output, format="plain"):
        if '' != cmd_output and cmd_output is not None:
            if "json" == format:
                columns = json.loads(cmd_output)
                for column in columns:
                    self.columns[column['Field']] = Column.new_from_dict(column)
            else:
                rows = cmd_output.split("\n")

                # Remove header row
                if rows[0].startswith("Field"):
                    rows.pop(0)

                for row in rows:
                    data = row.split("\t")
                    self.columns[data[0]] = Column.new_from_list(data)
                print(cmd_output)
        else:
            raise TypeError("cmd_output should not be an empty string or None.")

    def __getitem__(self, item):
        return self.columns[item]

    def __len__(self):
        return len(self.columns)

    def get_columns(self):
        return self.columns


class DBColumns(WPCommand):
    command = ['db', 'columns']

    # Render output in a particular format.
    # default: var_export
    # options: var_export, json, yaml
    format = ''

    columns = Columns

    def __init__(self, table, **args):
        super().__init__(**args)
        self.table = table

        self.format = self.get_arg_value(key="format", default_value=self.format)

    def params(self):
        return [
            self.table
        ]

    def get_excluded_attrs(self):
        return [
            "table",
            "columns",
        ]

    def run(self):
        out, err = super().run()
        if "" != out:
            self.columns = Columns(out)
        return out, err



from wp.cache_get import CacheGet


# Abstract base for cache incr and cache decr.
class CacheDecrIncr(CacheGet):
    command = ['cache']

    # [<offset>]
    # The amount by which to decrement the item’s value.
    # ---
    # default: 1
    # ---
    offset = 1

    def __init__(self, key, **args):
        super().__init__(key, **args)

        self.offset = self.get_arg_value(key="offset", default_value=self.offset)

    def params(self):
        return [
            self.key,
            self.offset,
            self.group,
        ]

    def get_excluded_attrs(self):
        return super().get_excluded_attrs() + [
            "offset",
        ]



from wp.command import WPCommand


class DBImport(WPCommand):
    command = ['db', 'import']

    # File.
    file = ''  # Required

    # When using an SQL file, do not include speed optimization such as disabling auto-commit and key checks.
    skip_optimization = bool

    # Extra arguments to pass to mysqldump. Refer to mysqldump docs.
    fields = []

    # Loads the environment’s MySQL option files.
    # Default behavior is to skip loading them to avoid failures due to misconfiguration.
    defaults = []

    def __init__(self, file, **args):
        super().__init__(**args)
        self.file = file

        self.skip_optimization = self.get_arg_value(key="skip_optimization", default_value=self.skip_optimization)
        self.fields = self.get_arg_value(key="fields", default_value=self.fields)
        self.defaults = self.get_arg_value(key="defaults", default_value=self.defaults)

    def params(self):
        return [
            self.file
        ]

    def get_excluded_attrs(self):
        return [
            "file"
        ]




from wp.cache_decr_incr import CacheDecrIncr


# TODO Test.
# Decrements a value in the object cache.
# Errors if the value can’t be decremented.
# wp cache decr my_key 2 my_group
# 48
class CacheDecr(CacheDecrIncr):
    command = ['cache', 'decr']



from wp.command import WPCommand

# Open /wp-admin/ in a browser.
# wp admin
class Admin(WPCommand):
    command = ['admin']

    def __init__(self, **args):
        super().__init__(**args)

    def params(self):
        return []



from wp.config import WPCLIConfig
# from wp.wp import WP, wp

# search-replace
from wp.search_replace import SearchReplace

# admin
from wp.admin import Admin

# cache
from wp.cache_get import CacheGet
from wp.cache_set import CacheSet
from wp.cache_add import CacheAdd
from wp.cache_flush import CacheFlush
from wp.cache_decr import CacheDecr
from wp.cache_incr import CacheIncr
from wp.cache_replace import CacheReplace
from wp.cache_type import CacheType

# eval
from wp.eval import Eval

# db
from wp.db_prefix import DBPrefix
from wp.db_query import DBQuery
from wp.db_check import DBCheck
from wp.db_import import DBImport
from wp.db_export import DBExport
from wp.db_search import DBSearch
from wp.db_columns import DBColumns
from wp.db_tables import DBTables
from wp.db_items import DBItems, DBItem

# post
from wp.post_get import PostGet
from wp.post_update import PostUpdate

# option
from wp.option_add import OptionAdd
from wp.option_get import OptionGet
from wp.option_update import OptionUpdate




from wp.command import WPCommand


# TODO Test.
# Base to a couple cap commands.
class Cap(WPCommand):
    command = ['cap']

    # <role>
    # Key for the role.
    role = ""

    # <cap>…
    # One or more capabilities to add.
    cap = ""

    def __init__(self, role, cap, **args):
        super().__init__(**args)

        self.value = role
        self.value = cap

    def params(self):
        return [
            self.role,
            self.cap,
        ]

    def get_excluded_attrs(self):
        return [
            "role",
            "cap",
        ]



from wp.command import WPCommand


class PostGet(WPCommand):
    command = ['post', 'get']

    # Post ID.
    id = int

    # Instead of returning the whole post, returns the value of a single field.
    field = ''

    # Limit the output to specific fields. Defaults to all fields.
    fields = ''

    # Render output in a particular format.
    # default: table
    # options: table, csv, json, yaml
    format = ''

    def __init__(self, id, **args):
        super().__init__(**args)
        self.id = id

        self.format = self.get_arg_value(key="format", default_value=self.format)
        self.field = self.get_arg_value(key="field", default_value=self.field)
        self.fields = self.get_arg_value(key="fields", default_value=self.fields)

    def params(self):
        return [
            str(self.id)
        ]

    def get_excluded_attrs(self):
        return [
            "id"
        ]

    # Gets the custom string attrs.
    def get_attr_custom_param_output(self, attr):
        if 'fields' == attr and hasattr(self, 'fields') and '' != self.fields:
            return '--fields=' + self.fields

        return ''



from wp.command import WPCommand


class SearchReplace(WPCommand):
    command = ['search-replace']

    # A string to search for within the database.
    old = ''  # Required

    # Replace instances of the first string with this new string.
    new = ''  # Required

    # List of database tables to restrict the replacement to. Wildcards are supported, e.g. 'wp_*options' or 'wp_post*'.
    table = ''  # Optional

    # Parameters
    # Run the entire search/replace operation and show report, but don’t save changes to the database.
    dry_run = bool

    # Search/replace through all the tables registered to $wpdb in a multisite install.
    network = bool

    # Enable replacement on any tables that match the table prefix even if not registered on $wpdb.
    all_tables_with_prefix = bool

    # Enable replacement on ALL tables in the database, regardless of the prefix, and even if not registered on $wpdb.
    # Overrides –network and –all-tables-with-prefix.
    all_tables = bool

    # Write transformed data as SQL file instead of saving replacements to the database.
    # If <file> is not supplied, will output to STDOUT.
    export = ''

    # Define number of rows in single INSERT statement when doing SQL export.
    # You might want to change this depending on your database configuration (e.g. if you need to do fewer queries).
    # Default: 50
    export_insert_size = int

    # Do not perform the replacement on specific tables. Use commas to specify multiple tables.
    # Wildcards are supported, e.g. 'wp_*options' or 'wp_post*'.
    skip_tables = ''

    # Do not perform the replacement on specific columns. Use commas to specify multiple columns.
    skip_columns = ''

    # Perform the replacement on specific columns. Use commas to specify multiple columns.
    include_columns = ''

    # Force the use of PHP (instead of SQL) which is more thorough, but slower.
    precise = bool

    # Enable recursing into objects to replace strings. Defaults to true; pass –no-recurse-objects to disable.
    recurse_objects = bool

    # Prints rows to the console as they’re updated.
    verbose = bool

    # Runs the search using a regular expression (without delimiters).
    # Warning: search-replace will take about 15-20x longer when using –regex.
    regex = bool

    # Pass PCRE modifiers to regex search-replace (e.g. ‘i’ for case-insensitivity).
    regex_flags = ''

    # The delimiter to use for the regex. It must be escaped if it appears in the search string.
    # The default value is the result of chr(1).
    regex_delimiter = ''

    # The maximum possible replacements for the regex per row (or per unserialized data bit per row).
    # Defaults to -1 (no limit).
    regex_limit = int

    # Render output in a particular format. Options: table, count
    format = ''

    # Produce report. Defaults to true.
    report = bool

    # Report changed fields only. Defaults to false, unless logging, when it defaults to true.
    report_changed_only = ''

    # Log the items changed. If <file> is not supplied or is “-“, will output to STDOUT.
    # Warning: causes a significant slow down, similar or worse to enabling –precise or –regex.
    log = ''

    # For logging, number of characters to display before the old match and the new replacement.
    # Default 40. Ignored if not logging.
    before_context = int

    # For logging, number of characters to display after the old match and the new replacement.
    # Default 40. Ignored if not logging.
    after_context = int

    def __init__(self, old, new, **args):
        super().__init__(**args)
        # super().__init__(self, self.command, **args)
        self.old = old
        self.new = new

        self.table = self.get_arg_value(key="table", default_value=self.table)
        self.dry_run = self.get_arg_value(key="dry_run", default_value=self.dry_run)
        self.network = self.get_arg_value(key="network", default_value=self.network)
        self.all_tables_with_prefix = self.get_arg_value(key="all_tables_with_prefix",
                                                         default_value=self.all_tables_with_prefix)
        self.all_tables = self.get_arg_value(key="all_tables", default_value=self.all_tables)
        self.export = self.get_arg_value(key="export", default_value=self.export)
        self.export_insert_size = self.get_arg_value(key="export_insert_size", default_value=self.export_insert_size)
        self.skip_tables = self.get_arg_value(key="skip_tables", default_value=self.skip_tables)
        self.skip_columns = self.get_arg_value(key="skip_columns", default_value=self.skip_columns)
        self.include_columns = self.get_arg_value(key="include_columns", default_value=self.include_columns)
        self.precise = self.get_arg_value(key="precise", default_value=self.precise)
        self.recurse_objects = self.get_arg_value(key="recurse_objects", default_value=self.recurse_objects)
        self.verbose = self.get_arg_value(key="verbose", default_value=self.verbose)
        self.regex = self.get_arg_value(key="regex", default_value=self.regex)
        self.regex_flags = self.get_arg_value(key="regex_flags", default_value=self.regex_flags)
        self.regex_delimiter = self.get_arg_value(key="regex_delimiter", default_value=self.regex_delimiter)
        self.regex_limit = self.get_arg_value(key="regex_limit", default_value=self.regex_limit)
        self.format = self.get_arg_value(key="format", default_value=self.format)
        self.report = self.get_arg_value(key="report", default_value=self.report)
        self.report_changed_only = self.get_arg_value(key="report_changed_only", default_value=self.report_changed_only)
        self.log = self.get_arg_value(key="log", default_value=self.log)
        self.before_context = self.get_arg_value(key="before_context", default_value=self.before_context)
        self.after_context = self.get_arg_value(key="after_context", default_value=self.after_context)

    def params(self):
        p = [
            self.old,
            self.new
        ]
        if "" != self.table:
            p.append(self.table)

        return p
        # return [
        #     self.old,
        #     self.new,
        #     self.table,
        #     # f"\"{self.old}\"",
        #     # f"\"{self.new}\"",
        #     # f"'{self.table}'"
        # ]

    def get_raw_params(self):
        return [
            "before_context",
            "after_context"
        ]

    def get_excluded_attrs(self):
        return [
            "old",
            "new",
            "table"
        ]

    def get_attr_custom_param_output(self, attr):
        if 'regex_flags' == attr and '' != self.regex_flags:
            return '--regex-flags=' + self.regex_flags
        if 'include_columns' == attr and '' != self.include_columns:
            return '--include-columns=' + self.include_columns

        return ''



from wp.command import WPCommand


class DBPrefix(WPCommand):
    command = ['db', 'prefix']

    def __init__(self, **args):
        super().__init__(**args)

    def params(self):
        return []



import yaml
import re


class WPCLIBaseItem(object):
    url = ''
    path = ''
    disabled_commands = []
    require = []

    def __init__(self, **args):
        self.url = args.get('url')
        self.path = args.get('path')
        self.disabled_commands = args.get('disabled_commands')
        self.require = args.get('require')

    def __add__(self, other):
        item = self
        item.maybe_set("url", other.url)
        item.maybe_set("path", other.path)
        item.maybe_set("disabled_commands", other.disabled_commands)
        item.maybe_set("require", other.require)

        return item

    def maybe_set(self, key, value):
        if self.has_value(value):
            super().__setattr__(key, value)

    @staticmethod
    def has_value(val):
        return val is not None or val != ''


class WPCLISSHItem(WPCLIBaseItem):
    cmd = ''

    def __init__(self, **args):
        super().__init__(**args)
        self.cmd = args.get('cmd')

    def __add__(self, other):
        item = super().__add__(other)
        item.maybe_set("cmd", other.cmd)

        return item


class WPCLISSH(object):
    vagrant = WPCLISSHItem
    development = WPCLISSHItem
    staging = WPCLISSHItem
    production = WPCLISSHItem

    def __init__(self, **args):
        self.vagrant = args.get('vagrant')
        self.development = args.get('development')
        self.staging = args.get('staging')
        self.production = args.get('production')

    def __add__(self, other):
        item = super().__add__(other)

        item.maybe_set("development", other.development)
        item.maybe_set("staging", other.staging)
        item.maybe_set("production", other.production)

        return item


class WPCLIItem(WPCLIBaseItem):
    ssh = WPCLISSH
    http = ''
    user = ''
    color = ''

    def __init__(self, **args):
        super().__init__(**args)
        self.ssh = args.get('ssh')
        self.http = args.get('http')
        self.user = args.get('user')
        self.color = args.get('color')

    def __add__(self, other):
        item = super().__add__(other)
        item.maybe_set("ssh", other.ssh)
        item.maybe_set("http", other.http)
        item.maybe_set("user", other.user)
        item.maybe_set("color", other.color)

        return item


class WPCLIConfig(WPCLIItem):
    ssh = WPCLISSH

    # Environments
    both = []
    development = WPCLIItem
    staging = WPCLIItem
    production = WPCLIItem
    vagrant = WPCLIItem

    # Global Parameters
    skip_themes = False
    skip_plugins = False
    skip_packages = False
    debug = False

    # Private Parameters
    yml_path = ''
    yml_str = ''

    def __init__(self, **args):
        super().__init__(**args)

        self.debug = args.get('debug')
        self.skip_themes = args.get('skip_themes')
        self.skip_plugins = args.get('skip_plugins')
        self.skip_packages = args.get('skip_packages')

        self.yml_path = args.get('yml_path')
        if self.yml_path is not None:
            self.load()

    def __add__(self, other):
        item = super().__add__(other)

        item.maybe_set("ssh", other.ssh)
        item.maybe_set("debug", other.debug)
        item.maybe_set("skip_themes", other.skip_themes)
        item.maybe_set("skip_plugins", other.skip_plugins)
        item.maybe_set("skip_packages", other.skip_packages)
        item.maybe_set("both", other.both)
        item.maybe_set("development", other.development)
        item.maybe_set("production", other.production)
        item.maybe_set("staging", other.staging)
        item.maybe_set("vagrant", other.vagrant)

        return item

    def normalize_yml(self):
        str = self.yml_str

        # Replace all ` - @`
        str = re.sub(
            r"^(\s-\s?)@",
            "\\1",
            str,
            flags=re.MULTILINE
        )

        # Replace all `^@`
        str = re.sub(
            r"^@",
            "",
            str,
            flags=re.MULTILINE
        )

        return "!WPCLIConfig\n" + str

    def load(self, path=''):
        if '' == path:
            path = self.yml_path

        try:
            # Load the wp-cli.sample.yml file into a string.
            with open(path, "r") as f:
                self.yml_str = f.read()
                f.close()
            self.yml_str = self.normalize_yml()

            # Parse the YML file.
            loader = self.get_loader()
            self._data = yaml.load(self.yml_str, Loader=loader)

            # Merge everything.
            self = self + self._data

        except yaml.YAMLError as e:
            print(e)

    def get_loader(self):
        """Add constructors to PyYAML loader."""
        loader = yaml.SafeLoader
        loader.add_constructor("!WPCLIConfig", wpcli_conf_constructor)
        return loader


def wpcli_conf_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> WPCLIConfig:
    return WPCLIConfig(**loader.construct_mapping(node))



from wp.command import WPCommand


class OptionAddUpdate(WPCommand):
    command = ['option', 'upsert']

    # Option key.
    key = ''

    # Option value.
    value = ''

    # Autoload: either yes/no.
    autoload = bool

    # Render output in a particular format.
    # default: plaintext
    # options: plaintext, json
    format = ''

    def __init__(self, key, value='', **args):
        super().__init__(**args)
        self.key = key
        self.value = value

        self.format = self.get_arg_value(key="format", default_value=self.format)

    def params(self):
        return [
            self.key,
            self.value,
        ]

    def get_excluded_attrs(self):
        return [
            "key",
            "value",
        ]



from wp.command import WPCommand


# Gets a value from the object cache.
# Errors if the value doesn’t exist.
# wp cache get my_key my_group
class CacheGet(WPCommand):
    command = ['cache', 'get']

    # <key>
    # Cache key.
    key = ""

    # [<group>]
    # Method for grouping data within the cache which allows the same key to be used across groups.
    # ---
    # default: default
    # ---
    group = ""

    def __init__(self, key, **args):
        super().__init__(**args)

        self.key = key
        self.group = self.get_arg_value(key="group", default_value=self.group)

    def params(self):
        return [
            self.key,
            self.group,
        ]

    def get_excluded_attrs(self):
        return [
            "key",
            "group",
        ]



from wp.option_add_update import OptionAddUpdate


class OptionUpdate(OptionAddUpdate):
    command = ['option', 'update']

    def __init__(self, key, value='', **args):
        super().__init__(key, value, **args)

    def params(self):
        return super().params()

    def get_excluded_attrs(self):
        return super().get_excluded_attrs()




from wp.command import WPCommand


class OptionGet(WPCommand):
    command = ['option', 'get']

    # Option key.
    key = ''

    # Render output in a particular format.
    # default: var_export
    # options: var_export, json, yaml
    format = ''

    def __init__(self, key, **args):
        super().__init__(**args)
        self.key = key

        self.format = self.get_arg_value(key="format", default_value=self.format)

    def params(self):
        return [
            str(self.key)
        ]

    def get_excluded_attrs(self):
        return [
            "key"
        ]



from wp.cache_get import CacheGet


# TODO Test.
# Sets a value to the object cache, regardless of whether it already exists.
# Errors if the value can’t be set.
# wp cache set my_key my_value my_group 300
# Success: Set object 'my_key' in group 'my_group'.
class CacheSet(CacheGet):
    command = ['cache', "set"]

    # <value>
    # Value to add to the key.
    value = ""

    # [<expiration>]
    # Define how long to keep the value, in seconds. 0 means as long as possible.
    # ---
    # default: 0
    # ---
    expiration = 0

    def __init__(self, key, value, **args):
        super().__init__(key, **args)

        self.value = value
        self.expiration = self.get_arg_value(key="expiration", default_value=self.expiration)

    def params(self):
        return [
            self.key,
            self.value,
            self.group,
            self.expiration,
        ]

    def get_excluded_attrs(self):
        return super().get_excluded_attrs() + [
            "value",
            "expiration",
        ]



from wp.command import WPCommand


class DBExport(WPCommand):
    command = ['db', 'export']

    # File.
    file = ''  # Required

    # add-drop-table
    add_drop_table = False

    # no-create-info
    no_create_info = False

    # Extra arguments to pass to mysqldump. Refer to mysqldump docs.
    fields = []

    # WHERE phrase
    where = ''

    # The comma separated list of specific tables to export.
    # Excluding this parameter will export all tables in the database.
    tables = ''

    # The comma separated list of specific tables that should be skipped from exporting.
    # Excluding this parameter will export all tables in the database.
    exclude_tables = ''

    # include-tablespaces: Skips adding the default –no-tablespaces option to mysqldump.
    include_tablespaces = ''

    # Output filename for the exported database.
    porcelain = ''

    # Loads the environment’s MySQL option files.
    # Default behavior is to skip loading them to avoid failures due to misconfiguration.
    defaults = []

    def __init__(self, file, **args):
        super().__init__(**args)
        self.file = file

        self.no_create_info = self.get_arg_value(key="no_create_info", default_value=self.no_create_info)
        self.add_drop_table = self.get_arg_value(key="add_drop_table", default_value=self.add_drop_table)
        self.where = self.get_arg_value(key="where", default_value=self.where)
        self.fields = self.get_arg_value(key="fields", default_value=self.fields)
        self.tables = self.get_arg_value(key="tables", default_value=self.tables)
        self.exclude_tables = self.get_arg_value(key="exclude_tables", default_value=self.exclude_tables)
        self.include_tablespaces = self.get_arg_value(key="include_tablespaces", default_value=self.include_tablespaces)
        self.porcelain = self.get_arg_value(key="porcelain", default_value=self.porcelain)
        self.defaults = self.get_arg_value(key="defaults", default_value=self.defaults)

        # Prevent stupidity.
        if self.add_drop_table:
            self.no_create_info = False

    def params(self):
        return [
            self.file
        ]

    def get_raw_params(self):
        return [
            "exclude_tables",
        ]

    def get_excluded_attrs(self):
        return [
            "file"
        ]

    # Gets the false attrs.
    def get_custom_bool_attrs(self):
        return {
            "no_create_info-true": "--no-create-info=true"
        }



from wp.command import WPCommand


class DBQuery(WPCommand):
    command = ['db', 'query']

    # SQL.
    sql = ''  # Required

    # Skips outputting column names.
    skip_column_names = bool

    def __init__(self, sql, **args):
        super().__init__(**args)
        self.sql = sql

        self.skip_column_names = self.get_arg_value(key="skip_column_names", default_value=self.skip_column_names)

    def params(self):
        return [
            self.sql
            # f"\"{self.sql}\""
        ]

    def get_excluded_attrs(self):
        return [
            "sql"
        ]

    @staticmethod
    def escape(str):
        return str.translate(
            str.maketrans({
                "\0": "\\0",
                "\r": "\\r",
                "\x08": "\\b",
                "\x09": "\\t",
                "\x1a": "\\z",
                "\n": "\\n",
                "\r": "\\r",
                "\"": "\\\"",
                "'": "\\'",
                "\\": "\\\\",
                "%": "\\%"
            }))



from wp.cache_decr_incr import CacheDecrIncr


# TODO Test.
# Increments a value in the object cache.
# wp cache incr my_key 2 my_group
# 50
class CacheIncr(CacheDecrIncr):
    command = ['cache', 'incr']



from wp.command import WPCommand


class Eval(WPCommand):
    command = ['eval']

    # <php-code>
    # The code to execute, as a string.

    # [--skip-wordpress]
    # Execute code without loading WordPress.
    skip_wordpress = bool

    def __init__(self, code, **args):
        super().__init__(**args)

        self.code = code
        self.skip_wordpress = self.get_arg_value(key="skip_wordpress", default_value=self.skip_wordpress)

    def params(self):
        return [
            self.code
        ]

    def get_excluded_attrs(self):
        return [
            "code"
        ]
